/*
 * Copyright (C) 2014-2019 Muhammad Tayyab Akram
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <SBConfig.h>
#include <SheenBidi.h>

#ifdef SB_CONFIG_UNITY

#include "BidiChain.c"
#include "BidiTypeLookup.c"
#include "BracketQueue.c"
#include "GeneralCategoryLookup.c"
#include "IsolatingRun.c"
#include "LevelRun.c"
#include "PairingLookup.c"
#include "RunQueue.c"
#include "SBAlgorithm.c"
#include "SBBase.c"
#include "SBCodepointSequence.c"
#include "SBLine.c"
#include "SBLog.c"
#include "SBMirrorLocator.c"
#include "SBParagraph.c"
#include "SBScriptLocator.c"
#include "ScriptLookup.c"
#include "ScriptStack.c"
#include "StatusStack.c"

#endif
