 //  SuperTuxKart - a fun racing game with go-kart
//  Copyright (C) 2009-2015 Marianne Gagnon
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 3
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#ifndef __HEADER_HELP_COMMON_HPP__
#define __HEADER_HELP_COMMON_HPP__

// This file contains include headers that are used by all or most help screens.
// It also contains a standalone function to switch between the help screens.
// This simplifies maintenance.

// Frequent widgets used by multiple option screens
#include "guiengine/widgets/ribbon_widget.hpp"

// Other help screens, for navigation between them
#include "states_screens/help/help_screen_1.hpp"
#include "states_screens/help/help_screen_2.hpp"
#include "states_screens/help/help_screen_3.hpp"
#include "states_screens/help/help_screen_4.hpp"
#include "states_screens/help/help_screen_5.hpp"
#include "states_screens/help/help_screen_6.hpp"
#include "states_screens/help/help_screen_7.hpp"

// GUI management
#include "states_screens/state_manager.hpp"

namespace HelpCommon
{
	void switchTab(std::string selected_tab);
}

#endif