//  SuperTuxKart - A fun racing game with go-karts
//  Copyright (C) 2017-18 QwertyChouskie
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 3
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#ifndef HEADER_LINK_HELPER_HPP
#define HEADER_LINK_HELPER_HPP

#include <string>

#include <IEventReceiver.h>
namespace irr
{
    namespace gui
    {
        class IGUIStaticText;
    }
}


namespace Online
{
    class LinkHelper
    {
    public:
        static bool isSupported();
        static void openURL(const std::string& url);
        static bool openURLIrrElement(irr::gui::IGUIStaticText* text, irr::SEvent::SMouseInput mouse);
    }; //class LinkHelper
} // namespace Online
#endif // link_helper_HPP
